from gyver.database.typedef import Driver

from .interface import Dialect
from .mysql import MariaDbDialect
from .mysql import MysqlDialect
from .postgres import PostgresDialect
from .sqlite import SqliteDriver


def resolve_driver(driver: Driver) -> Dialect:
    _table: dict[Driver, Dialect] = {
        Driver.MYSQL: MysqlDialect(),
        Driver.POSTGRES: PostgresDialect(),
        Driver.SQLITE: SqliteDriver(),
        Driver.MARIADB: MariaDbDialect(),
    }
    return _table[driver]


def build_dialect_scheme(dialect: Dialect, sync: bool = False):
    return "+".join(
        item
        for item in (
            dialect.dialect_name,
            dialect.sync_driver if sync else dialect.async_driver,
        )
        if item
    )



from gyver.model import Model
from gyver.utils import to_camel


class BotoModel(Model):
    class Config:
        @staticmethod
        def alias_generator(field: str):
            return field[:1].upper() + to_camel(field[1:])



from typing import Literal
from typing import Optional
from typing import overload

from gyver.utils import lazyfield

from .config import CacheConfig
from .interface import AsyncCacheInterface
from .mapper import AsyncCacheMap
from .mock import MockAsyncCache
from .redis import AsyncRedisWrapper


class AsyncCacheProvider:
    @overload
    def __init__(self, *, config: CacheConfig):
        ...

    @overload
    def __init__(self, *, test: Literal[True]):
        ...

    def __init__(
        self, *, config: Optional[CacheConfig] = None, test: bool = False
    ) -> None:
        self._config = config
        self._test = test

    @staticmethod
    def _make_concrete(
        config: Optional[CacheConfig], test: bool
    ) -> AsyncCacheInterface:
        return MockAsyncCache() if test else AsyncRedisWrapper(config)

    @lazyfield
    def interface(self):
        return self._make_concrete(self._config, self._test)

    def mapper(self, name: str) -> AsyncCacheMap:
        return AsyncCacheMap(self.interface, name)



from gyver.config import ProviderConfig


class QueueConfig(ProviderConfig):
    __prefix__ = "queue"

    name: str
    message_batch_limit: int = 10
    visibility_timeout: int = 10
    wait_time_seconds: int = 3
    stream_sleep_seconds: float = 0.1



import typing

from context_handler.ext.sqlalchemy import AsyncSaAdapter
from context_handler.ext.sqlalchemy import SaAdapter

from gyver.config import from_config

from .config import DatabaseConfig
from .config import DefaultDatabaseConfig
from .utils import make_uri


class SyncDatabaseProvider(SaAdapter):
    def __init__(self, config: typing.Optional[DatabaseConfig] = None) -> None:
        self._config = config or from_config(DefaultDatabaseConfig)
        super().__init__(uri=make_uri(self.config, sync=True))

    @property
    def config(self):
        return self._config


class AsyncDatabaseProvider(AsyncSaAdapter):
    def __init__(self, config: typing.Optional[DatabaseConfig] = None) -> None:
        self._config = config or from_config(DefaultDatabaseConfig)
        super().__init__(uri=make_uri(self.config))

    @property
    def config(self):
        return self._config



import typing
from abc import ABC
from abc import abstractmethod

from sqlalchemy.sql import Select

from gyver.database.typedef import ClauseType
from gyver.utils.helpers import cache

from . import comp as cp
from .interface import ApplyClause
from .interface import Comparator
from .where import Where


class Paginate(ApplyClause, ABC):
    type_ = ClauseType.APPLY

    @typing.final
    def __init__(self, limit: int, offset: int) -> None:
        self._limit = limit
        self._offset = offset

    @abstractmethod
    def apply(self, query: Select) -> Select:
        raise NotImplementedError

    @staticmethod
    def none():
        return _NullPaginate(limit=0, offset=0)

    def __bool__(self):
        return isinstance(self, _NullPaginate)


class LimitOffsetPaginate(Paginate):
    def apply(self, query: Select) -> Select:
        return query.limit(self._limit).offset(self._offset)


@cache
def make_field_paginate(
    field: str, jump_comparison: Comparator = cp.greater
) -> type[Paginate]:
    class FieldPaginate(Paginate):
        def apply(self, query: Select) -> Select:
            return query.where(
                Where(field, self._offset, jump_comparison).bind(
                    query.selected_columns  # type: ignore
                )
            ).limit(self._limit)

    return FieldPaginate


IdPaginate = make_field_paginate("id")


class _NullPaginate(Paginate):
    def apply(self, query: Select) -> Select:
        return query



import typing
from datetime import datetime

from gyver.utils import json

T = typing.TypeVar("T")


class AsyncCacheInterface(typing.Protocol):
    async def get(
        self, name: str, cast: typing.Callable[[typing.Any], T] = json.loads
    ) -> T:
        ...

    async def set(
        self,
        name: str,
        value: typing.Any,
        expires_at: typing.Union[datetime, typing.Literal[-1]],
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        ...

    async def map_get(
        self,
        map_name: str,
        name: str,
        cast: typing.Callable[[typing.Any], T] = json.loads,
    ) -> T:
        ...

    async def map_set(
        self,
        map_name: str,
        name: str,
        value: typing.Any,
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        ...

    async def delete(
        self, name: str, map_name: typing.Optional[str] = None
    ) -> None:
        ...


class CacheInterface(typing.Protocol):
    def get(
        self, name: str, cast: typing.Callable[[typing.Any], T] = json.loads
    ) -> T:
        ...

    def set(
        self,
        name: str,
        value: typing.Any,
        expires_at: typing.Union[datetime, typing.Literal[-1]],
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        ...

    def map_get(
        self,
        map_name: str,
        name: str,
        cast: typing.Callable[[typing.Any], T] = json.loads,
    ) -> T:
        ...

    def map_set(
        self,
        map_name: str,
        name: str,
        value: typing.Any,
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        ...

    def delete(self, name: str, map_name: typing.Optional[str] = None) -> None:
        ...



import contextlib
from inspect import isclass

from typing_extensions import TypeGuard

from gyver.database.entity import AbstractEntity

from .exc import FieldNotFound
from .interface import FieldType
from .interface import Mapper


def retrieve_attr(entity: Mapper, field: str) -> FieldType:
    is_entity = _is_entity(entity)
    if field == "id" and is_entity:
        field = "id_"
    try:
        with contextlib.suppress(AttributeError):
            return getattr(entity, field)
        return getattr(entity.c, field)  # type: ignore
    except AttributeError:
        name = entity.__name__ if is_entity else "query"
        raise FieldNotFound(name, field) from None


def _is_entity(entity: Mapper) -> TypeGuard[type[AbstractEntity]]:
    return isclass(entity) and issubclass(entity, AbstractEntity)



import importlib
import inspect
import pathlib
import typing
from dataclasses import dataclass

from gyver.exc import GyverError

from .lazy import lazyfield

T = typing.TypeVar("T")

Validator = typing.Callable[[typing.Any], bool]
StrOrPath = typing.Union[str, pathlib.Path]
_convert_path = pathlib.Path.as_posix


class InvalidPath(GyverError, ValueError):
    pass


@dataclass(frozen=True)
class Finder:

    validator: Validator
    root: pathlib.Path
    look_on: typing.Optional[pathlib.Path] = None
    exclude: typing.Sequence[StrOrPath] = ()

    def __post_init__(self):
        if not self.root.exists():
            raise InvalidPath(
                f"root must be a valid path, received {self.root}"
            )

    @property
    def _look_on(self):
        return self.look_on or self.root

    @lazyfield
    def _exclude(self):
        return *self.exclude, "__init__.py", "__pycache__"

    @lazyfield
    def output(self) -> dict[str, typing.Any]:
        return {}

    def find(self):
        if self._look_on.is_dir():
            self._iterate_folder(self._look_on)
        else:
            self._iterate_file_contents(self._look_on)

    def _make_module_name(self, path: pathlib.Path):
        target_str = _convert_path(path)
        return (
            target_str.replace(_convert_path(self.root), self.root.name)
            .removesuffix(".py")
            .replace("/", ".")
        )

    def _should_look(self, path: pathlib.Path):
        str_exclude = tuple(
            item for item in self._exclude if isinstance(item, str)
        )
        path_exclude = tuple(
            item for item in self._exclude if isinstance(item, pathlib.Path)
        )
        return (
            path.name not in str_exclude
            and path not in path_exclude
            and (not path.is_file() or path.name.endswith(".py"))
        )

    def _iterate_folder(self, path: pathlib.Path):
        for item in path.iterdir():
            if not self._should_look(item):
                continue
            if item.is_dir():
                self._iterate_folder(item)
            else:
                self._iterate_file_contents(item)

    def _iterate_file_contents(self, path: pathlib.Path):
        mod = importlib.import_module(self._make_module_name(path))
        for name, obj in inspect.getmembers(mod):
            if self.validator(obj):
                self.output[name] = obj


class MissingParams(GyverError, NotImplementedError):
    pass


def class_validator(*types: type):
    if not types:
        raise MissingParams("Missing types in class_validator paramaters")

    def validator(obj: typing.Any):
        return (
            issubclass(obj, types) and obj not in types
            if inspect.isclass(obj)
            else False
        )

    return validator


def instance_validator(*types: type):
    if not types:
        raise MissingParams("Missing types in instance_validator paramaters")

    def validator(obj: typing.Any):
        return isinstance(obj, types)

    return validator



from contextlib import suppress
from typing import Any
from typing import Iterable
from typing import TypeVar
from typing import get_args
from typing import get_origin

from pydantic.fields import ModelField

from gyver.config.utils import boolean_cast
from gyver.exc import MissingName
from gyver.model import Model
from gyver.utils.exc import panic
from gyver.utils.strings import make_lex_separator

from .config import MISSING
from .config import Config


class ProviderConfig(Model):
    __prefix__ = ""
    __without_prefix__ = ()

    class Config:
        alias_generator = str.upper
        fields = {"exclude": {}}


ProviderT = TypeVar("ProviderT", bound=ProviderConfig)

_default_config = Config()


def from_config(
    provider: type[ProviderT],
    *,
    __config__: Config = _default_config,
    **presets: Any,
) -> ProviderT:

    fields = tuple(
        field
        for field in _get_fields(provider)
        if field.name not in (*presets, "__prefix__", "__without_prefix__")
    )
    result = {
        field.alias: _get_value(provider, field, __config__)
        for field in fields
    }
    return provider.parse_obj(result | presets)


def _get_fields(provider: type[ProviderT]) -> Iterable[ModelField]:
    return provider.__fields__.values()


def _get_value(
    provider: type[ProviderConfig], field: ModelField, config: Config
):
    name, alias = _make_names(provider, field)
    default = _get_default(field)
    cast = _get_cast(field)
    return _tryeach(alias, name, default=default, cast=cast, config=config)


def _tryeach(*names: str, default: Any, cast: Any, config: Config) -> Any:
    for name in names:
        with suppress(MissingName):
            return config(name, cast, default)
    raise panic(
        MissingName, f"{','.join(names)} not found and no default was given"
    )


def _make_names(
    provider: type[ProviderConfig], field: ModelField
) -> tuple[str, str]:
    name = field.name
    alias = field.alias
    if not provider.__prefix__ or {name, alias}.intersection(
        provider.__without_prefix__
    ):
        return (name, alias)
    prefix = provider.__prefix__.removesuffix("_")
    name = name.removeprefix("_")
    alias = alias.removeprefix("_")
    return f"{prefix}_{name}".upper(), f"{prefix}_{alias}".upper()


def _get_cast(field: ModelField):
    _sequences = (list, tuple, set)
    outer_type = field.outer_type_
    origin = get_origin(outer_type)
    if outer_type is bool:
        return boolean_cast
    if not origin:
        return (
            make_lex_separator(outer_type)
            if outer_type in _sequences
            else outer_type
        )
    if (origin := get_origin(outer_type)) in _sequences:
        assert origin is not None
        args = get_args(outer_type)
        cast = args[0] if args else str
        return make_lex_separator(origin, cast)
    if dict in (origin, outer_type):
        raise NotImplementedError("Unable to handle dict-like variables")
    return origin


def _get_default(field: ModelField):
    if field.default not in (None, Ellipsis):
        return field.default
    return (
        MISSING if field.default_factory is None else field.default_factory()
    )



from .config import MISSING
from .config import Config
from .config import EnvMapping
from .envconfig import DotFile
from .envconfig import EnvConfig
from .provider import ProviderConfig
from .provider import from_config
from .typedef import Env
from .utils import boolean_cast

__all__ = [
    "boolean_cast",
    "Config",
    "EnvMapping",
    "MISSING",
    "ProviderConfig",
    "from_config",
    "EnvConfig",
    "Env",
    "DotFile",
]



from contextlib import suppress
from typing import Any

from orjson import JSONDecodeError

from gyver.boto.models import BotoModel
from gyver.utils import json


class Message(BotoModel):
    message_id: str
    receipt_handle: str
    body: str


class Messages(BotoModel):
    messages: list[Message]
    response_metadata: dict[str, Any]


class MessageWrapper:
    def __init__(self, message: Message) -> None:
        self._message = message

    def get(self):
        return self._message

    def loads(self):
        with suppress(JSONDecodeError):
            return json.loads(self._message.body)



from typing import Protocol

from gyver.database.typedef import Driver


class Dialect(Protocol):
    default_port: int
    driver: Driver
    dialect_name: str
    async_driver: str
    sync_driver: str
    only_host: bool



import typing

T = typing.TypeVar("T")


def make_singleton(cls: type[T]) -> type[T]:
    instance = None
    init_called = False
    cls_new = cls.__new__
    cls_init = cls.__init__

    def custom_new(*args, **kwds) -> T:
        nonlocal instance
        if instance is None:
            new_instance = (
                object.__new__(cls)
                if cls_new is object.__new__
                else cls_new(*args, **kwds)
            )
            instance = new_instance
        return instance

    def custom_init(*args, **kwds):
        nonlocal init_called
        if not init_called:
            cls_init(*args, **kwds)
            init_called = True

    cls.__new__ = custom_new  # type: ignore
    cls.__init__ = custom_init

    return cls



import logging
from typing import BinaryIO
from typing import Optional
from typing import Protocol

from botocore.exceptions import ClientError

from gyver.boto.base import AWSCredentials
from gyver.boto.base import AWSCredentialsConfig
from gyver.boto.base import make_client
from gyver.config import ProviderConfig
from gyver.config import from_config
from gyver.exc import GyverError
from gyver.utils import lazyfield


class StorageConfig(ProviderConfig):
    __prefix__ = "storage"

    bucket_name: str
    presigned_expiration: int = 3600


class FileNotFound(GyverError, FileNotFoundError):
    pass


class ReadableBuffer(Protocol):
    def read(self) -> bytes:
        ...


class StorageProvider:
    def __init__(
        self,
        config: Optional[StorageConfig] = None,
        aws_credentials: Optional[AWSCredentialsConfig] = None,
    ) -> None:
        self._config = config or from_config(StorageConfig)
        self._aws_credentials = aws_credentials

    @lazyfield
    def client(self):
        return make_client("s3", self._aws_credentials)

    @property
    def credentials(self):
        return self._aws_credentials or AWSCredentials().credentials

    @property
    def config(self):
        return self._config

    def _generate_presigned_url(
        self,
        method: str,
        object_name: str,
        bucket_name: Optional[str] = None,
    ):
        bucket_name = bucket_name or self._config.bucket_name
        try:
            response = self.client.generate_presigned_url(
                method,
                Params={"Bucket": bucket_name, "Key": object_name},
                ExpiresIn=self._config.presigned_expiration,
            )
        except ClientError as e:
            logging.error(e)
            return object_name
        else:
            return response

    def upload_file(
        self,
        buf: BinaryIO,
        object_name: str,
        bucket_name: Optional[str] = None,
    ):
        bucket_name = bucket_name or self._config.bucket_name
        self.client.upload_fileobj(buf, bucket_name, object_name)

    def generate_presigned_get(
        self, object_name: str, bucket_name: Optional[str] = None
    ):
        bucket_name = bucket_name or self._config.bucket_name
        return (
            self._generate_presigned_url(
                "get_object", object_name, bucket_name
            )
            if self.is_key(object_name, bucket_name)
            else object_name
        )

    def is_key(
        self, object_name: str, bucket_name: Optional[str] = None
    ) -> bool:
        bucket_name = bucket_name or self._config.bucket_name
        try:
            self.client.head_object(Bucket=bucket_name, Key=object_name)
        except ClientError:
            return False
        else:
            return True

    def get_file(
        self, object_name: str, bucket_name: Optional[str] = None
    ) -> ReadableBuffer:
        bucket_name = bucket_name or self._config.bucket_name
        if not self.is_key(object_name, bucket_name):
            raise FileNotFound(
                f"Object {object_name} not found in {bucket_name} bucket"
            )
        response = self.client.get_object(Bucket=bucket_name, Key=object_name)
        return response["Body"]



from . import json
from . import timezone
from .exc import panic
from .finder import Finder
from .finder import class_validator
from .finder import instance_validator
from .helpers import cache
from .helpers import frozen
from .lazy import lazyfield
from .singleton import make_singleton
from .strings import to_camel
from .strings import to_snake
from .strings import upper_camel

__all__ = [
    "lazyfield",
    "to_camel",
    "to_snake",
    "upper_camel",
    "json",
    "timezone",
    "frozen",
    "cache",
    "Finder",
    "class_validator",
    "instance_validator",
    "make_singleton",
    "panic",
]



from gyver.database.typedef import Driver
from gyver.utils import frozen


@frozen
class SqliteDriver:
    default_port = 0
    driver = Driver.SQLITE
    dialect_name = "sqlite"
    async_driver = "aiosqlite"
    sync_driver = ""
    only_host = True



import typing

import sqlalchemy as sa
from sqlalchemy.sql import Delete
from sqlalchemy.sql import Select
from sqlalchemy.sql import Update

from gyver.utils import lazyfield

from . import _helpers
from ._helpers import retrieve_attr
from .interface import ApplyClause
from .interface import BindClause
from .interface import Mapper

ExecutableType = typing.Union[Select, Update, Delete]
ExecutableT = typing.TypeVar("ExecutableT", bound=ExecutableType)
Self = typing.TypeVar("Self", bound="Query")


class Query(typing.Generic[ExecutableT]):
    def __init__(
        self,
        entity: Mapper,
        executable: typing.Callable[..., ExecutableT],
    ) -> None:
        self._entity = entity
        self._executable = executable

    def _preprocess_params(
        self, entity: Mapper
    ) -> typing.Sequence[typing.Any]:
        return (entity,)

    def _process_query(self, query: ExecutableT) -> ExecutableT:
        return query

    @lazyfield
    def query(self):
        return self._process_query(
            self._executable(*self._preprocess_params(self._entity))
        )

    def bind(self: Self, *where: BindClause) -> Self:
        self.query = _process_filters(self.query, self._entity, where)
        return self

    def get(self):
        return self.query


class select(Query[Select]):
    def __init__(
        self,
        entity: Mapper,
        apply_func: typing.Optional[
            typing.Callable[[Mapper], tuple[typing.Any, ...]]
        ] = None,
    ) -> None:
        super().__init__(entity, sa.select)
        self._apply_func = apply_func or self._default_apply

    @staticmethod
    def _default_apply(entity: Mapper) -> tuple[Mapper]:
        return (entity,)

    def _preprocess_params(
        self, entity: Mapper
    ) -> typing.Sequence[typing.Any]:
        return self._apply_func(entity)

    def apply(self: Self, *clauses: ApplyClause) -> Self:
        for item in clauses:
            self.query = item.apply(self.query)
        return self


class update(Query[Update]):
    def __init__(
        self, entity: Mapper, values: typing.Mapping[str, typing.Any]
    ) -> None:
        super().__init__(entity, sa.update)
        self._values = values

    def _process_query(self, query: Update) -> Update:
        return query.values(self._values)


class delete(Query[Delete]):
    def __init__(self, entity: Mapper) -> None:
        super().__init__(entity, sa.delete)


def count(entity: Mapper, field: str = "id"):
    def _make_count(entity: Mapper):
        return (sa.func.count(retrieve_attr(entity, field)),)

    return select(entity, _make_count)


def distinct(entity: Mapper, *fields: str):
    def _make_distinct(ent: Mapper):
        return tuple(
            _helpers.retrieve_attr(ent, field).distinct() for field in fields
        )

    return select(entity, _make_distinct)


def _process_filters(
    query: ExecutableType, entity: Mapper, where: typing.Sequence
) -> ExecutableType:
    return query.where(*(clause.bind(entity) for clause in where))



from .base import AWSCredentials
from .base import AWSCredentialsConfig
from .simple_queue import QueueConfig
from .simple_queue import QueueProvider
from .simple_queue import QueueStream
from .storage import StorageConfig
from .storage import StorageProvider

__all__ = [
    "AWSCredentialsConfig",
    "AWSCredentials",
    "StorageConfig",
    "StorageProvider",
    "QueueStream",
    "QueueConfig",
    "QueueProvider",
]



from . import comp
from ._helpers import retrieve_attr
from .builder import delete
from .builder import select
from .builder import update
from .exc import FieldNotFound
from .exc import FilterError
from .interface import ApplyClause
from .interface import BindClause
from .interface import Mapper
from .order_by import OrderBy
from .order_by import OrderDirection
from .paginate import IdPaginate
from .paginate import LimitOffsetPaginate
from .paginate import Paginate
from .paginate import make_field_paginate
from .utils import as_date
from .utils import as_lower
from .utils import as_time
from .utils import as_upper
from .where import Where
from .where import and_
from .where import or_

__all__ = [
    "Where",
    "and_",
    "or_",
    "OrderBy",
    "OrderDirection",
    "as_date",
    "as_time",
    "as_lower",
    "as_upper",
    "IdPaginate",
    "LimitOffsetPaginate",
    "Paginate",
    "make_field_paginate",
    "FilterError",
    "FieldNotFound",
    "comp",
    "select",
    "update",
    "delete",
    "BindClause",
    "ApplyClause",
    "retrieve_attr",
    "Mapper",
]



from gyver.exc import GyverError


class FilterError(GyverError):
    """Base Error for all filter related exceptions"""


class FieldNotFound(FilterError, AttributeError):
    def __init__(self, name: str, field: str) -> None:
        super().__init__(f"type {name} has no {field} attribute")



from typing import TYPE_CHECKING
from typing import Literal
from typing import Optional
from typing import Union
from typing import overload

from boto3 import Session

from gyver.config import ProviderConfig
from gyver.config import from_config
from gyver.utils import make_singleton

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_sqs.client import SQSClient


class AWSCredentialsConfig(ProviderConfig):
    __prefix__ = "aws"

    access_key_id: str
    secret_access_key: str
    region_name: str
    endpoint_url: Optional[str] = None


@make_singleton
class AWSCredentials:
    def __init__(self) -> None:
        self._aws_credentials = from_config(AWSCredentialsConfig)

    @property
    def credentials(self):
        return self._aws_credentials


@overload
def make_client(
    service_name: Literal["s3"],
    aws_credentials: Optional[AWSCredentialsConfig] = None,
) -> "S3Client":
    ...


@overload
def make_client(
    service_name: Literal["sqs"],
    aws_credentials: Optional[AWSCredentialsConfig] = None,
) -> "SQSClient":
    ...


def make_client(
    service_name: Literal["s3", "sqs"],
    aws_credentials: Optional[AWSCredentialsConfig] = None,
) -> Union["S3Client", "SQSClient"]:
    aws_credentials = aws_credentials or AWSCredentials().credentials
    kwargs = {
        "region_name": aws_credentials.region_name,
        "aws_access_key_id": aws_credentials.access_key_id,
        "aws_secret_access_key": aws_credentials.secret_access_key,
    }
    if (endpoint_url := aws_credentials.endpoint_url) is not None:
        kwargs["endpoint_url"] = endpoint_url
    return Session().client(service_name, **kwargs)



from types import FunctionType
from typing import Any
from typing import Optional
from typing import Sequence
from typing import Union

from gyver import utils
from gyver.config import ProviderConfig
from gyver.exc import InvalidField
from gyver.utils.exc import panic

from . import drivers
from .typedef import Driver


class DatabaseConfig(ProviderConfig):
    driver: Driver
    host: str
    port: int = -1
    user: str = ""
    password: str = ""
    name: str = ""

    @utils.lazyfield
    def real_port(self) -> int:
        return self.port if self.port != -1 else self.dialect.default_port

    @utils.lazyfield
    def dialect(self) -> drivers.Dialect:
        if self.driver is Driver.CUSTOM and not self.resolved:
            raise panic(
                InvalidField,
                'Field driver of value "custom" missing custom dialect,'
                "dont forget to call .override_dialect(dialect)",
            )
        return drivers.resolve_driver(self.driver)

    @utils.lazyfield
    def resolved(self):
        return False

    def override_dialect(self, dialect: drivers.Dialect) -> None:
        self.dialect = dialect
        self.resolved = True


def make_database_config(
    prefix: str,
    without_prefix: Sequence[str] = (),
    name: str = "DatabaseConfig",
    module: Optional[str] = __name__,
    **defaults: Union[Any, FunctionType],
) -> type[DatabaseConfig]:  # sourcery skip: instance-method-first-arg-name
    new_cls = type(
        name,
        (DatabaseConfig,),
        {
            "__prefix__": prefix,
            "__without_prefix__": without_prefix,
            "__module__": module,
        },
    )
    for key, value in defaults.items():
        _override_defaults(new_cls, key, value)
    return new_cls


def _override_defaults(
    new_cls: type[DatabaseConfig], key: str, value: Union[Any, FunctionType]
):
    try:
        modelfield = new_cls.__fields__[key]
    except KeyError:
        raise panic(
            InvalidField,
            f"Field {key} not found in model {new_cls.__name__}",
        ) from None
    else:
        if isinstance(value, FunctionType):
            modelfield.default_factory = value
        else:
            modelfield.default


DefaultDatabaseConfig = make_database_config("db")



from typing import Any

from pydantic import BaseModel

from gyver import utils


class Model(BaseModel):
    class Config:
        json_loads = utils.json.loads
        json_dumps = utils.json.dumps
        frozen = True
        orm_mode = True
        alias_generator = utils.to_camel
        allow_population_by_field_name = True
        keep_untouched = (utils.lazyfield,)

    def __setattr__(self, name: str, value: Any):
        if isinstance(getattr(type(self), name, None), utils.lazyfield):
            object.__setattr__(self, name, value)
            return
        return super().__setattr__(name, value)


class MutableModel(BaseModel):
    class Config:
        frozen = False



from gyver.database.typedef import Driver
from gyver.utils import frozen


@frozen
class MysqlDialect:
    default_port = 3306
    driver = Driver.MYSQL
    dialect_name = "mysql"
    async_driver = "aiomysql"
    sync_driver = "pymysql"
    only_host = False


class AsyncMyDialect(MysqlDialect):
    async_driver = "asyncmy"


class MariaDbDialect(MysqlDialect):
    dialect_name = "mariadb"


class AsyncMyMariaDialect(AsyncMyDialect, MariaDbDialect):
    pass



import typing

from gyver.utils import json

from .interface import AsyncCacheInterface
from .interface import CacheInterface

T = typing.TypeVar("T")


class AsyncCacheMap:
    def __init__(self, cache: AsyncCacheInterface, name: str) -> None:
        self._cache = cache
        self._name = name

    async def get(
        self, name: str, cast: typing.Callable[[typing.Any], T] = json.loads
    ) -> T:
        return await self._cache.map_get(self._name, name, cast)

    async def set(
        self,
        name: str,
        value: typing.Any,
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        return await self._cache.map_set(self._name, name, value, dumps)

    async def delete(self, name: str):
        return await self._cache.delete(name, self._name)


class CacheMap:
    def __init__(self, cache: CacheInterface, name: str) -> None:
        self._cache = cache
        self._name = name

    def get(
        self, name: str, cast: typing.Callable[[typing.Any], T] = json.loads
    ) -> T:
        return self._cache.map_get(self._name, name, cast)

    def set(
        self,
        name: str,
        value: typing.Any,
        dumps: typing.Callable[[typing.Any], str] = json.dumps,
    ) -> None:
        return self._cache.map_set(self._name, name, value, dumps)

    def delete(self, name: str):
        return self._cache.delete(name, self._name)



from urllib.parse import quote

from .config import CacheConfig


def make_uri(config: CacheConfig) -> str:
    return (
        (
            f"redis://{quote(config.user)}:{quote(config.password)}"
            f"@{config.host}:{config.port}/"
        )
        if all((config.user, config.password))
        else f"redis://{config.host}:{config.port}"
    )



from . import query
from .config import DefaultDatabaseConfig
from .config import make_database_config
from .drivers import Dialect
from .entity import AbstractEntity
from .entity import Entity
from .entity import make_table
from .metadata import metadata as default_metadata
from .provider import AsyncDatabaseProvider
from .provider import SyncDatabaseProvider
from .typedef import Driver
from .utils import make_uri

__all__ = [
    "Dialect",
    "DefaultDatabaseConfig",
    "make_database_config",
    "Driver",
    "make_uri",
    "SyncDatabaseProvider",
    "AsyncDatabaseProvider",
    "Entity",
    "AbstractEntity",
    "make_table",
    "query",
    "default_metadata",
]



from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import overload

from boto3 import Session

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_sqs.client import SQSClient


@overload
def make_client(service_name: Literal["s3"], **params) -> "S3Client":
    ...


@overload
def make_client(service_name: Literal["sqs"], **params) -> "SQSClient":
    ...


def make_client(service_name: Literal["s3", "sqs"], **params: Any) -> Any:
    return Session().client(service_name, **params)



from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Callable
from typing import Literal
from typing import Optional
from typing import TypeVar
from typing import Union

import redis
from redis import asyncio as aioredis

from gyver.config import from_config
from gyver.exc import CacheMiss
from gyver.utils import json
from gyver.utils import lazyfield

from .config import CacheConfig
from .interface import AsyncCacheInterface
from .interface import CacheInterface
from .utils import make_uri

T = TypeVar("T")


class AsyncRedisWrapper(AsyncCacheInterface):
    def __init__(self, cache_config: Optional[CacheConfig] = None) -> None:
        self._config = cache_config or from_config(CacheConfig)

    def _pool(self):
        return aioredis.ConnectionPool.from_url(make_uri(self._config))

    @lazyfield
    def _instance(self):
        return aioredis.Redis(connection_pool=self._pool())

    async def get(self, name: str, cast: Callable[[Any], T] = json.loads) -> T:
        response = await self._instance.get(name)
        if response is None:
            raise CacheMiss
        return cast(response)

    async def set(
        self,
        name: str,
        value: Any,
        expires_at: Union[datetime, Literal[-1]],
        dumps: Callable[[Any], str] = json.dumps,
    ) -> None:
        expiration = None if expires_at == -1 else expires_at
        await self._instance.set(
            name,
            dumps(value),
            ex=expiration and expiration - datetime.now(timezone.utc),
        )

    async def map_get(
        self, map_name: str, name: str, cast: Callable[[Any], T] = json.loads
    ) -> T:
        try:
            response = await self._instance.hget(map_name, name)
        except aioredis.RedisError:
            raise CacheMiss from None
        else:
            if response is None:
                raise CacheMiss
            return cast(response)

    async def map_set(
        self,
        map_name: str,
        name: str,
        value: Any,
        dumps: Callable[[Any], str] = json.dumps,
    ) -> None:
        await self._instance.hset(
            map_name,
            name,
            dumps(value),
        )

    async def delete(self, name: str, map_name: Optional[str] = None) -> None:
        if map_name is None:
            await self._instance.delete(name)
        else:
            await self._instance.hdel(map_name, name)


class RedisWrapper(CacheInterface):
    def __init__(self, cache_config: Optional[CacheConfig] = None) -> None:
        self._config = cache_config or from_config(CacheConfig)

    def _pool(self):
        return redis.ConnectionPool.from_url(make_uri(self._config))

    @lazyfield
    def _instance(self):
        return redis.Redis(connection_pool=self._pool())

    def get(self, name: str, cast: Callable[[Any], T] = json.loads) -> T:
        response = self._instance.get(name)
        if response is None:
            raise CacheMiss
        return cast(response.decode("utf-8"))

    def set(
        self,
        name: str,
        value: Any,
        expires_at: Union[datetime, Literal[-1]],
        dumps: Callable[[Any], str] = json.dumps,
    ) -> None:
        expiration = None if expires_at == -1 else expires_at
        self._instance.set(
            name,
            dumps(value),
            ex=expiration and expiration - datetime.now(timezone.utc),
        )

    def map_get(
        self, map_name: str, name: str, cast: Callable[[Any], T] = json.loads
    ) -> T:
        try:
            response = self._instance.hget(map_name, name)
        except aioredis.RedisError:
            raise CacheMiss from None
        else:
            if response is None:
                raise CacheMiss
            return cast(response.decode("utf-8"))

    def map_set(
        self,
        map_name: str,
        name: str,
        value: Any,
        dumps: Callable[[Any], str] = json.dumps,
    ) -> None:
        self._instance.hset(
            map_name,
            name,
            dumps(value),
        )

    def delete(self, name: str, map_name: Optional[str] = None) -> None:
        if map_name is None:
            self._instance.delete(name)
        else:
            self._instance.hdel(map_name, name)



import time
from typing import Any
from typing import Callable
from typing import Generator
from typing import Generic
from typing import Optional
from typing import TypeVar

from botocore.exceptions import ClientError

from gyver.boto.base import AWSCredentials
from gyver.boto.base import AWSCredentialsConfig
from gyver.boto.base import make_client
from gyver.config import from_config
from gyver.exc import QueueNotFound
from gyver.utils import lazyfield
from gyver.utils import panic

from .config import QueueConfig
from .models import Message
from .models import Messages
from .models import MessageWrapper

T = TypeVar("T")


class QueueProvider:
    def __init__(
        self,
        config: Optional[QueueConfig] = None,
        aws_credentials: Optional[AWSCredentialsConfig] = None,
    ) -> None:
        self._config = config or from_config(QueueConfig)
        self._aws_credentials = aws_credentials

    @lazyfield
    def client(self):
        return make_client("sqs", self._aws_credentials)

    @property
    def config(self):
        return self._config

    @property
    def aws_credentials(self):
        return self._aws_credentials or AWSCredentials()

    @lazyfield
    def queue_url(self):
        try:
            queue_url = self.client.get_queue_url(QueueName=self._config.name)
        except ClientError:
            raise panic(
                QueueNotFound,
                f"{self._config.name} does not exist on given credentials",
            ) from None
        else:
            return queue_url["QueueUrl"]

    def get_messages(self):
        response = self.client.receive_message(
            QueueUrl=self.queue_url,
            MaxNumberOfMessages=self._config.message_batch_limit,
            VisibilityTimeout=self._config.visibility_timeout,
            WaitTimeSeconds=self._config.wait_time_seconds,
        )
        return Messages.parse_obj(response)

    def send_message(self, body: str, **kwargs: Any):
        self.client.send_message(
            QueueUrl=self.queue_url, MessageBody=body, **kwargs
        )

    def stream(
        self,
        cast: Callable[[Message], T] = MessageWrapper,
        max_rounds: int = 0,
    ) -> "QueueStream[T]":
        return QueueStream(self, cast, max_rounds)


class QueueStream(Generic[T]):
    def __init__(
        self,
        provider: QueueProvider,
        cast: Callable[[Message], T],
        max_rounds: int,
    ) -> None:
        self._internal_provider = provider
        self._cast = cast
        self._max_rounds = max_rounds

    def _iter(self) -> Generator[T, None, None]:
        counter = 0
        while True:
            message_response = self._internal_provider.get_messages()
            yield from (self._cast(item) for item in message_response.messages)
            time.sleep(self._internal_provider.config.stream_sleep_seconds)
            counter += 1
            if self._max_rounds > 0 and self._max_rounds == counter:
                break

    def __iter__(self):
        yield from self._iter()



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Union

from gyver import utils

from .config import Config
from .config import EnvMapping
from .config import default_mapping
from .typedef import Env


@dataclass(frozen=True)
class DotFile:
    filename: Union[str, Path]
    env: Env
    apply_to_lower: bool = False

    def __gt__(self, other: "DotFile") -> bool:
        return self.env.weight > other.env.weight

    def is_higher(self, env: Env) -> bool:
        return self.env.weight >= env.weight


class EnvConfig(Config):
    def __init__(
        self,
        *dotfiles: DotFile,
        env_var: str = "CONFIG_ENV",
        mapping: EnvMapping = default_mapping
    ) -> None:
        self._env_var = env_var
        self._dotfiles = dotfiles
        self._mapping = mapping
        self._file_values = {}
        if self.dotfile:
            self._read_file(self.dotfile.filename)

    @utils.lazyfield
    def env(self):
        return self.__call__(self._env_var, Env.new)

    @utils.lazyfield
    def dotfile(self):
        for dot in sorted(self._dotfiles, reverse=True):
            if not dot.is_higher(self.env):
                break
            if dot.env != self.env and (
                not dot.apply_to_lower or not dot.is_higher(self.env)
            ):
                continue
            if not os.path.isfile(dot.filename):
                continue
            return dot



import typing

from typing_extensions import Self

T = typing.TypeVar("T")
SelfT = typing.TypeVar("SelfT")


class lazyfield(typing.Generic[SelfT, T]):
    def __init__(self, func: typing.Callable[[SelfT], T]) -> None:
        self._func = func

    def __set_name__(self, owner: type[SelfT], name: str):
        self.public_name = name
        self.private_name = f"_lazyfield_{name}"

    @typing.overload
    def __get__(self, instance: SelfT, owner: type[SelfT]) -> T:
        ...

    @typing.overload
    def __get__(
        self, instance: typing.Literal[None], owner: type[SelfT]
    ) -> Self:
        ...

    def __get__(
        self,
        instance: typing.Optional[SelfT],
        owner: typing.Optional[type[SelfT]] = None,
    ) -> typing.Union[T, Self]:
        if not instance:
            return self
        assert instance is not None
        try:
            val = typing.cast(
                T,
                object.__getattribute__(
                    instance,
                    self.private_name,
                ),
            )
        except AttributeError:
            val = self._try_set(instance)
        return val

    def _try_set(self, instance: SelfT) -> T:
        try:
            val = self._func(instance)
        except Exception as e:
            # remove exception context to create easier traceback
            raise e from None
        else:
            object.__setattr__(instance, self.private_name, val)
            return val

    def __set__(self, instance: SelfT, value: T):
        object.__setattr__(instance, self.private_name, value)

    def __delete__(self, instance: SelfT):
        object.__delattr__(instance, self.private_name)



class GyverError(Exception):
    """Base exception for all gyver exceptions"""


class InvalidCast(GyverError):
    """Exception raised for config when
    cast callable raises an error"""


class MissingName(GyverError, KeyError):
    """Exception raised for config when
    name is not found in given environ"""


class InvalidField(GyverError, KeyError):
    """Exception raised when field lookup
    found nothing on mapping"""


class CacheMiss(GyverError, KeyError):
    """Exception raised when cache lookup
    returns None or raises error"""


class QueueNotFound(InvalidField):
    """Exception raised when queue name provider
    does not exist on given credentials"""



from urllib.parse import quote

from gyver.utils import cache

from . import drivers
from .config import DatabaseConfig


@cache
def make_uri(config: DatabaseConfig, sync: bool = False) -> str:
    if config.dialect.only_host:
        return f"{drivers.build_dialect_scheme(config.dialect, sync)}://{config.host}"
    return (
        f"{drivers.build_dialect_scheme(config.dialect, sync)}://"
        f"{quote(config.user)}:{quote(config.password)}@"
        f"{config.host}:{config.real_port}/"
        f"{quote(config.name)}"
    )



from gyver.config import ProviderConfig


class CacheConfig(ProviderConfig):
    __prefix__ = "cache"

    host: str
    port: int = 6379
    user: str = ""
    password: str = ""
    max_connections: int = 10
    default_key_expiration: int = -1



import typing

import sqlalchemy as sa

from gyver.database.typedef import ClauseType

from . import _helpers
from . import comp as cp
from . import interface

T = typing.TypeVar("T")


class Where(interface.BindClause, typing.Generic[T]):
    def __init__(
        self,
        field: str,
        expected: typing.Optional[T] = None,
        comp: interface.Comparator[T] = cp.equals,
    ) -> None:
        self.field = field
        self.expected = expected
        self.comp = comp

    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        attr = _helpers.retrieve_attr(mapper, self.field)
        return (
            sa.true()
            if self.expected is None
            else self.comp(attr, self.expected)
        )


class _JoinBind(interface.BindClause):
    def __init__(
        self,
        items: typing.Sequence[interface.BindClause],
        operator: typing.Callable[..., interface.Comparison],
    ) -> None:
        self.items = items
        self.operator = operator

    type_ = ClauseType.BIND

    def bind(self, mapper: interface.Mapper) -> interface.Comparison:
        return self.operator(*(item.bind(mapper) for item in self.items))


def and_(*bind: interface.BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.and_)


def or_(*bind: interface.BindClause) -> _JoinBind:
    return _JoinBind(bind, sa.or_)


and_(Where("anyfield", 1))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gyver',
 'gyver.boto',
 'gyver.boto.simple_queue',
 'gyver.cache',
 'gyver.config',
 'gyver.crypto',
 'gyver.database',
 'gyver.database.drivers',
 'gyver.database.query',
 'gyver.database.session.asgi',
 'gyver.utils']

package_data = \
{'': ['*']}

install_requires = \
['context-handler>=5.1.0,<6.0.0',
 'cryptography>=38.0.4,<39.0.0',
 'orjson>=3.8.1,<4.0.0',
 'pydantic[email]>=1.10.2,<2.0.0',
 'typing-extensions>=4.4.0,<5.0.0']

extras_require = \
{':sys_platform != "linux"': ['tzdata>=2022.6,<2023.0'],
 'cache': ['redis>=4.4.0,<5.0.0'],
 'db-mariadb': ['pymysql>=1.0.2,<2.0.0', 'asyncmy>=0.2.5,<0.3.0'],
 'db-mysql': ['aiomysql>=0.1.1,<0.2.0', 'pymysql>=1.0.2,<2.0.0'],
 'db-pg': ['psycopg2>=2.9.5,<3.0.0', 'asyncpg>=0.27.0,<0.28.0'],
 'db-sqlite': ['aiosqlite>=0.17.0,<0.18.0'],
 'sqlalchemy': ['sqlalchemy>=1.4.43,<2.0.0']}

setup_kwargs = {
    'name': 'gyver',
    'version': '0.9.0',
    'description': 'Toolbox for web development',
    'long_description': '\n# Gyver\n\nToolbox for web development\n\n\n## Authors\n\n- [@guscardvs](https://www.github.com/guscardvs)\n\n\n## Installation\n\nInstall gyver with pip\n\n```bash\n  pip install gyver\n```\n    \n## License\n\n[MIT](https://choosealicense.com/licenses/mit/)\n\n\n## Roadmap\n\n- Database helpers\n\n- Http helpers\n\n',
    'author': 'Gustavo Correa',
    'author_email': 'self.gustavocorrea@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/guscardvs/gyver',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import typing

from context_handler.ext.sqlalchemy import AsyncSaContext
from context_handler.ext.sqlalchemy import SaContext
from sanic import Request
from sanic import Sanic

from gyver.database.config import DatabaseConfig
from gyver.database.provider import AsyncDatabaseProvider
from gyver.database.provider import SyncDatabaseProvider

TrxOptions = typing.Optional[typing.Literal["open", "begin"]]


def setup_database_session(
    app: Sanic,
    config: typing.Optional[DatabaseConfig] = None,
    sync: bool = False,
):
    provider_class = SyncDatabaseProvider if sync else AsyncDatabaseProvider

    async def _setup(_):
        provider = provider_class(config)
        app.ctx.gyver_database = provider

    app.register_listener(_setup, "before_server_start")


def get_async_context(transaction_on: TrxOptions = "open"):
    def _get(request: Request) -> AsyncSaContext:
        provider = typing.cast(
            AsyncDatabaseProvider, request.app.ctx.gyver_database
        )
        return provider.context(transaction_on=transaction_on)

    return _get


def get_sync_context(transaction_on: TrxOptions = "open"):
    def _get(request: Request) -> SaContext:
        provider = typing.cast(
            SyncDatabaseProvider, request.app.ctx.gyver_database
        )
        return provider.context(transaction_on=transaction_on)

    return _get



from .config import QueueConfig
from .models import Message
from .models import Messages
from .provider import QueueProvider
from .provider import QueueStream

__all__ = [
    "QueueConfig",
    "QueueProvider",
    "QueueStream",
    "Message",
    "Messages",
]



import functools
from typing import Callable
from typing import TypeVar
from typing import cast

from typing_extensions import ParamSpec

from .exc import panic


def _not_implemented(msg: str):
    def inner(*_, **__kwds__):
        raise panic(NotImplementedError, msg)

    return inner


P = ParamSpec("P")
T = TypeVar("T")


def frozen(cls: type[T]) -> type[T]:
    """Removes setattr and delattr from class"""
    type.__setattr__(
        cls,
        "__setattr__",
        _not_implemented(f"{cls.__name__} is frozen, cannot change values"),
    )
    type.__setattr__(
        cls,
        "__delattr__",
        _not_implemented(f"{cls.__name__} is frozen, cannot delete values"),
    )

    return cls


def cache(f: Callable[P, T]) -> Callable[P, T]:
    return cast(Callable[P, T], functools.cache(f))



from gyver.database.typedef import Driver
from gyver.utils import frozen


@frozen
class PostgresDialect:
    default_port = 5432
    driver = Driver.POSTGRES
    dialect_name = "postgresql"
    async_driver = "asyncpg"
    sync_driver = "psycopg2"
    only_host = False
